'use strict';
const crypto = require('node:crypto');
const KoaRouter = require('koa-router');
const qrcode = require('qrcode');
const randomstring = require('randomstring');
const dbOps = require('../data/dbops.js');
const glob = require('../global');

const router = new KoaRouter();

const appId = 'iGSLUsFI2xOEoVrC';
const webName = 'Acme Corporation';

router
  .get('/', front)
  .get('/auth', frontOps)
  .get('/signedup', signedup)
  .get('/signedin', signedin)
  .post('/auth-signup', authenticatorSignup)
  .post('/auth-signin', authenticatorSignin)
  
// ***** Route functions *****
// *

// The base URL endpoint with the initial page
//
async function front(ctx) {
  await ctx.render('front');
}

async function frontOps(ctx) {
  if (ctx.query.op === 'signup') {
    let dataUrl = await qrcode.toDataURL('signup:'+appId+':'+webName);
    await ctx.render('qrsignup', {image: dataUrl});
  }
  if (ctx.query.op === 'signin') {
    let challenge = randomstring.generate(16);
    glob.challenge = challenge;
    let dataUrl = await qrcode.toDataURL('signin:'+appId+':'+webName+':'+challenge);
    await ctx.render('qrsignin', {image: dataUrl});
  }
}

async function signedup(ctx) {
  if (!!glob.user)
    await ctx.render('aftersignup', {uname: glob.user});
  else
    ctx.body = 'Error: Sign up failed!';
  glob.user = null;
}

async function signedin(ctx) {
  if (!!glob.user && !!glob.challenge)
    await ctx.render('aftersignin', {uname: glob.user, challenge: glob.challenge});
  else
    ctx.body = 'Error: User not authenticated!'
  glob.user = null;
  glob.challenge = null;
}

async function authenticatorSignup(ctx) {
  let result = '';
  console.log(ctx.request.body);
  const message = ctx.request.body.name + ctx.request.body.uid + ctx.request.body.auth + ctx.request.body.user;
  const verified = verifySign(ctx.request.body.auth, message, ctx.request.body.sig);  
  console.log("Signature is " + (verified ? 'valid' : 'invalid'));
  if (verified) {
    const user = {
      Uid: ctx.request.body.uid,
      Name: ctx.request.body.name,
      AuthCert: ctx.request.body.auth,
      UserCert: ctx.request.body.user
    }
    result = await dbOps.insertUser(user);
    if (result.lastID !== 0) {
      glob.user = ctx.request.body.name;
      glob.socket.send('signedup');
      ctx.body = 'ok - user registered as: ' + user.Name;
      return;
    }
  }
  glob.user = null;
  ctx.body = 'Invalid authenticator: ' + result;
}

async function authenticatorSignin(ctx) {
  console.log(ctx.request.body);
  console.log(ctx.headers);
  const user = await dbOps.getUserByUid(ctx.request.body.uid);
  if (user.Name && glob.challenge) {
    const message = ctx.request.body.uid + ctx.request.body.cha;
    const authVerified = verifySign(user.AuthCert, message, ctx.request.body.sig);
    const userVerified = verifySign(user.UserCert, glob.challenge, ctx.request.body.cha);
    console.log("Signature is " + (authVerified ? 'valid' : 'invalid'));
    console.log("User challenge is " + (userVerified ? 'verified' : 'not verified'));
    console.log("User: " + user.Name);
  
    if (authVerified && userVerified) {
      glob.user = user.Name;
      glob.socket.send('signedin');
      ctx.body = 'ok - user signed in: ' + user.Name;
      return;
    }
  }
  glob.challenge = null;
  ctx.body = 'User not authenticated!';
}

function verifySign(cert, message, sig) {
  const pubKey = crypto.createPublicKey('-----BEGIN CERTIFICATE-----\n' + cert + '-----END CERTIFICATE-----');
  const verifier = crypto.createVerify('SHA256');
  const msg = Buffer.from(message, 'utf-8');
  verifier.update(msg);
  return verifier.verify(pubKey, sig, 'base64');
}

// ***** export the router object *****
// *
module.exports = router;